@@ -1253,11 +1253,10 @@ impl JsRuntime {
   ) -> Result<ModuleId, AnyError> {
     let module_map_rc = Self::module_map(self.v8_isolate());
 
-    let load = module_map_rc.borrow().load_main(specifier.as_str(), code);
-
-    let (_load_id, prepare_result) = load.prepare().await;
-
-    let mut load = prepare_result?;
+    let mut load = module_map_rc
+      .borrow()
+      .load_main(specifier.as_str(), code)
+      .await?;
 
     while let Some(info_result) = load.next().await {
       let info = info_result?;
@@ -1268,7 +1267,8 @@ impl JsRuntime {
     }
 
     let root_id = load.expect_finished();
-    self.instantiate_module(root_id).map(|_| root_id)
+    self.instantiate_module(root_id)?;
+    Ok(root_id)
   }
 
   fn poll_pending_ops(
